/************************************************************************
     File:        CubeIt1.cpp

     Author:     
                  Michael Gleicher, gleicher@cs.wisc.edu
     Modifier
                  Yu-Chi Lai, yu-chi@cs.wisc.edu
     
     Comment:     Simple OpenGL / FlTk 3D Program
									originally written 11/10/99, Michael L. Gleicher
								 (it was called "3d.cpp")

                  Hacked 10/27/01 to use ArcBall instead of orbit cam
                  Major revision 10/11/2005 to clean up code

									This shows a simple program that creates a world and lets
									the user control the viewpoint using an ArcBall and move
									points around.

									It also provides "drop shadows" using the stencil buffer
									and the most basic UI

									This program provides a VERY simple interface for selecting points
									(click the left mouse button, selected point is cycled), and a
									really simple way to move the selected points around (using FlTk
									Widgets). 

									The purpose of this terrible interface is to show that 3D UIs don't
									need to be hard - you can do something workable and easy, allowing
									you to focus on getting the "real" work done

									This uses some simple utility code written for 559 projects
									(ArcBallCam and 3DUtils)

     In this file is:
									MyGlWindow - a class that is an OpenGL widget that
 															provides a view onto the world and handles
															the user interface
									main - a simple main function that creates a window
												with a MyGlWindow widget in it
   
     Platform:    Visio Studio.Net 2003 (upgraded to 2005)

     CVS Info: Date checked in: $Date: 2007/10/22 15:38:30 $
               Revision number: $Revision: 1.5 $

********************************************************************/
#include "stdio.h"
#include "math.h"
#include <windows.h>

// the FlTk headers have lots of warnings - these are bad, but there's not
// much we can do about them
#pragma warning(push)
#pragma warning(disable:4311)		// convert void* to long
#pragma warning(disable:4312)		// convert long to void*
#include <FL/Fl_Gl_Window.h>
#include <Fl/Fl.h>
#include <Fl/Fl_Multiline_Output.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <Fl/Fl_Double_Window.h>
#include <Fl/Fl_Button.h>
#include <FL/FL_Roller.H>
#pragma warning(pop)

#include "ArcBallCam.H"
#include "3DUtils.H"
#include "CubeIt.H"

//*********************************************************************************
//
// Cheesy UI!
// Need a callback for each dimension
// and a function to make the rollers
// Forward declare here, and put the code at the end
void makeRollers(int x, int y, int h, Fl_Widget* w);
void xRollerCB(Fl_Widget* w, void* p);
void yRollerCB(Fl_Widget* w, void* p);
void zRollerCB(Fl_Widget* w, void* p);


//*************************************************************************
//
// Create a widget that is a GL Window that has the world in it
//
// If you remember from the example program, all of the work happens
// (or can happen) in a Widget class that has draw and handle
// functions. In this example, I break those two "standard" routines
// into a bunch of pieces to make the code more organized
//
//*************************************************************************
class MyGlWindow : public Fl_Gl_Window {
	public:
		MyGlWindow(int x, int y, int w, int h);
	private:
		void draw();					// standard FlTk
		int handle(int);				// standard FlTk

		// The arc ball encapsulates the current state of the rotation
		// of the world inside of it (it actually stores it as a quaternion,
		// but we don't need to worry about that)
		ArcBallCam ball;

		// Put things to set up the projection into a method so we can use
		// it places other than draw
		void setProjection();
};

//**************************************************************************
//
// Implementation of MyGLWindow
//==========================================================================
MyGlWindow::MyGlWindow(int x, int y, int w, int h) 
:  Fl_Gl_Window(x,y,w,h)
//==========================================================================
{
	// note: we will use some fancy GL features, so we have to set the
	// mode correctly
	// specifically, we need the stencil buffer (so we can draw shadows
	// only on the ground plane) and alpha blending (to have transparent
	// shadows)
 mode( FL_RGB|FL_ALPHA|FL_DOUBLE | FL_STENCIL );

 ball.setup(this);
}

//**************************************************************************
//
// draw the windows proceeds as follows:
// first, clear the window (be sure to clear the Z-buffer)
// then set up the projection and view
// then draw the ground plane
// then set up for shadows and draw the objects
//		which causes their shadows to be drawn
// then draw the objects
//==========================================================================
void MyGlWindow::draw(void)
//==========================================================================
{

  // clear the window, be sure to clear the Z-Buffer too
  glClearColor(0,0,.3f,0);		// background should be blue

  // we need to clear out the stencil buffer since we'll use
  // it for shadows
  glClearStencil(0);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    
  // set up the drawing transforms
  setProjection();

  // now draw the ground plane
  setupFloor();
  drawFloor();

  // now to draw the shadows
  setupShadows();

  // Draw the objects without setting the colors (otherwise we'd
  // get colored shadows
  drawCubes(0);

  unsetupShadows();

  // Finally, draw the objects
  drawCubes();
}

//**************************************************************************
//
// * Set up the view projection matrix which is useful for selection
//==========================================================================
void MyGlWindow::setProjection(void)
//==========================================================================
{
	ball.setProjection();
}

//**************************************************************************
//
// Handle whatever events come our way
//
// The mouse behavior is:
//    right mouse = control view. Basically, it orbits around the center
//    (if we're looking in 3D)
//    we'll remember where the mouse goes down for later
//==========================================================================
int MyGlWindow::handle(int e)
//==========================================================================
{
	if (ball.handle(e)) return 1;
  switch(e) {
		case FL_SHOW:		// you must handle this, or not be seen!
			show();
			return 1;
		case FL_PUSH:		// any mouse button (since the arcball got the right button)
			selectedCube ++;
			selectedCube = selectedCube % (int) theCubes.size();
			damage(1);
			return 1;
  }
  return 0;
}

// the main routine makes the window, and then runs an even loop
// until the window is closed
int main(int, char**)
{
	Fl_Double_Window* wind = new Fl_Double_Window(100,100,600,400,"GL 3D Sample");

	wind->begin();		// put widgets inside of the window
		MyGlWindow* gl = new MyGlWindow(10,10,380,380);
		makeRollers(440,10,100,gl);
		Fl_Multiline_Output* out = new Fl_Multiline_Output(420,140,580-420,380-140);
		out->value("CubeIt 1! - Basic 3D UI\n\nLeftMouse: select\nRightMouse: ArcBall\nALT+RightMouse:pan Camera\nRollers: move selected");
	wind->end();


	wind->show();	// this actually opens the window

	// make some cubes to show
	theCubes.push_back(Cube(-3,3, 3));
	theCubes.push_back(Cube( 3,3, 3));
	theCubes.push_back(Cube(-3,3,-3));
	theCubes.push_back(Cube( 3,3,-3));
	// put some cubes over space so we can test the shadows only cast
	// onto the floor (not onto space)
	theCubes.push_back(Cube(7,3,0,1));
	theCubes.push_back(Cube(-7,3,0,1));
	theCubes.push_back(Cube(0,3,7,1));
	theCubes.push_back(Cube(0,3,-7,1));
	// put a cube over another cube so we can see that the shadows
	// don't stack
	theCubes.push_back(Cube( 0,4,-4,2));
	theCubes.push_back(Cube( 0,2,-3,1));

	// become interactive
	Fl::run();
	delete wind;

	return 1;
}

// actual implementation of the cheesy UI

//*********************************************************************************
//
// Cheesy UI!
// Need a callback for each dimension
// the reseting each time doesn't really work, but its a cheesy
// way to avoid having to reset the rollers when a new object is selected
//
//*********************************************************************************

//*********************************************************************************
//
// * Callback for X dimension
//=================================================================================
void xRollerCB(Fl_Widget* w, void* p)
//=================================================================================
{
  if (selectedCube > -1) {
	theCubes[selectedCube].x += (float) static_cast<Fl_Roller*>(w)->value();
	static_cast<Fl_Roller*>(w)->value(0);
	if (p) static_cast<Fl_Widget*>(p)->damage(1);
  }
}
//*********************************************************************************
//
// * Callback for Y dimension
//=================================================================================
void yRollerCB(Fl_Widget* w, void* p)
//=================================================================================
{
  if (selectedCube > -1) {
	theCubes[selectedCube].y += (float) static_cast<Fl_Roller*>(w)->value();
	static_cast<Fl_Roller*>(w)->value(0);
	if (p) static_cast<Fl_Widget*>(p)->damage(1);
  }
}
//*********************************************************************************
//
// * Callback for Z dimension
//=================================================================================
void zRollerCB(Fl_Widget* w, void* p)
//=================================================================================
{
  if (selectedCube > -1) {
	theCubes[selectedCube].z += (float) static_cast<Fl_Roller*>(w)->value();
	static_cast<Fl_Roller*>(w)->value(0);
	if (p) static_cast<Fl_Widget*>(p)->damage(1);
  }
}

//*********************************************************************************
// a cheesy UI
//
// Use FlTk Rollers to move the selected cube
//
// I don't recommend this interface - its just here to show you that
// there are simple things that work
//
// Make rollers - be sure you're inside of an FlTk Window
//                give it an xy position and size for the rollers
//                also give it an fltk widget to damage (your window)
//                anytime something changes
//*********************************************************************************
void makeRollers(int x, int y, int h, Fl_Widget* w)
{
  Fl_Roller* rx = new Fl_Roller(x,y,25,h,"X");
  rx->minimum(-1);
  rx->maximum(1);
  rx->callback(xRollerCB,w);
  Fl_Roller* ry = new Fl_Roller(x+40,y,25,h,"Y");
  ry->minimum(-1);
  ry->maximum(1);
  ry->callback(yRollerCB,w);
  Fl_Roller* rz = new Fl_Roller(x+80,y,25,h,"Z");
  rz->minimum(-1);
  rz->maximum(1);
  rz->callback(zRollerCB,w);
}


