/************************************************************************
     File:        CubeIt2.cpp

			October 2007 - updated to use newer ArcBallCam

     Author:     
                  Michael Gleicher, gleicher@cs.wisc.edu
     Modifier
                  Yu-Chi Lai, yu-chi@cs.wisc.edu
     
     Comment:     
									This shows a simple program that creates a world and lets
									the user control the viewpoint using an ArcBall and move
									points around.
		 
									It also provides "drop shadows" using the stencil buffer
									and the most basic UI

									This program provides a VERY simple interface for selecting points
									(click the left mouse button, selected point is cycled), and a
									really simple way to move the selected points around (using FlTk
									Widgets). 

									The purpose of this terrible interface is to show that 3D UIs don't
									need to be hard - you can do something workable and easy, allowing
									you to focus on getting the "real" work done

									This uses some simple utility code written for 559 projects
									(MyArcBall and 3DUtils)
     In this file is:
									MyGlWindow - a class that is an OpenGL widget that
 															provides a view onto the world and handles
															the user interface
									main - a simple main function that creates a window
												with a MyGlWindow widget in it

             Written at October 2005
  
     Platform:    Visio Studio.Net 2003 (updated to VS2005)

     CVS Info: Date checked in: $Date: 2007/10/22 15:38:30 $
               Revision number: $Revision: 1.5 $

*************************************************************************/

// the FlTk headers have lots of warnings - these are bad, but there's not
// much we can do about them
#pragma warning(push)
#pragma warning(disable:4311)		// convert void* to long
#pragma warning(disable:4312)		// convert long to void*
#include <FL/Fl_Gl_Window.h>
#include <Fl/Fl.h>
#include <Fl/Fl_Multiline_Output.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <Fl/Fl_Double_Window.h>
#include <Fl/Fl_Button.h>
#pragma warning(pop)

#include "stdio.h"
#include "math.h"

#include "ArcBallCam.H"
#include "3DUtils.H"
#include "CubeIt.H"

//********************************************************************
//
// Create a widget that is a GL Window that has the world in it
//
// if you remember from the example program, all of the work happens
// (or can happen) in a Widget class that has draw and handle
// functions. In this example, I break those two "standard" routines
// into a bunch of pieces to make the code more organized
//
//********************************************************************
class MyGlWindow : public Fl_Gl_Window {
	public:
		MyGlWindow(int x, int y, int w, int h);
	private:
		void draw();					// standard FlTk
		int handle(int);				// standard FlTk

		// use the arc ball camera interface
		ArcBallCam arcball;

		// put things to set up the projection into a method so we can use
		// it places other than draw
		// since we might want to put the projection onto the stack after a
		// pick matrix, provide an option to not clear the projection stack
		void setProjection(const bool clearProjection = true);

		// do the picking - make it a method so that its easier to read
		void doPick();

		// internal routine
		// this gets the position of the mouse in "normalized device coordinates"
		// that is X & Y go from -1 to 1 (with zero at the center)
		void getMouseNDC(float& x, float& y);
};

//**************************************************************************
//
// Implementation of MyGLWindow
//==========================================================================
MyGlWindow::MyGlWindow(int x, int y, int w, int h) :
  Fl_Gl_Window(x,y,w,h)
//==========================================================================
{
	// note: we will use some fancy GL features, so we have to set the
	// mode correctly
	// specifically, we need the stencil buffer (so we can draw shadows
	// only on the ground plane) and alpha blending (to have transparent
	// shadows)
 mode( FL_RGB|FL_ALPHA|FL_DOUBLE | FL_STENCIL );

 arcball.setup(this);
}

double m1x=0, m2x=0, m1y=0, m2y=5, m1z=5, m2z=5;

//**************************************************************************
//
// draw the windows proceeds as follows:
// first, clear the window (be sure to clear the Z-buffer)
// then set up the projection and view
// then draw the ground plane
// then set up for shadows and draw the objects
//		which causes their shadows to be drawn
// then draw the objects
//==========================================================================
void MyGlWindow::draw()
//==========================================================================
{
  // clear the window, be sure to clear the Z-Buffer too
  glClearColor(0,0,.3f,0);		// background should be blue
  // we need to clear out the stencil buffer since we'll use
  // it for shadows
  glClearStencil(0);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
  
  
  // set up the drawing transforms
  setProjection();

  // now draw the ground plane
  setupFloor();
  drawFloor();

  // now to draw the shadows
  setupShadows();
  // draw the objects without setting the colors (otherwise we'd
  // get colored shadows
  drawCubes(0);
  unsetupShadows();

  // finally, draw the objects
  drawCubes();
}

//**************************************************************************
//
// * Set up the view projection matrix which is useful for selection
//==========================================================================
void MyGlWindow::setProjection(const bool clearProjection)
//==========================================================================
{
  glMatrixMode(GL_PROJECTION);
  glViewport(0,0,w(),h());
  arcball.setProjection(clearProjection);
}

static int last_push;

//**************************************************************************
//
// Handle whatever events come our way
//
// The mouse behavior is:
//    right mouse = control view. Basically, it orbits around the center
//    (if we're looking in 3D)
//    we'll remember where the mouse goes down for later
//
// an issue: the fl event for dragging is the last button pressed 
// so using the shift key can mess you up - we'll remember what the
// last push event was, so we can do the right thing
//==========================================================================
int MyGlWindow::handle(int e)
//==========================================================================
{
  if (arcball.handle(e)) return 1;
  switch(e) {
  case FL_SHOW:		// you must handle this, or not be seen!
	  show();
	  return 1;
  case FL_PUSH:
	  switch(Fl::event_button()) {
		case 1:  // left mouse button down is "select" - do picking
			doPick();
			last_push = 1;
		  break;
	  };
	  damage(1);
	  return 1;
  case FL_RELEASE:
	  damage(1);
	  return 1;
  case FL_DRAG: // if the user drags the mouse
	  switch(last_push /*Fl::event_button()*/) {
	  case 1: // drag with the left mouse
		  if (selectedCube > -1) {
		  double r1x, r1y, r1z, r2x, r2y, r2z;
		  getMouseLine(r1x,r1y,r1z, r2x,r2y,r2z);

		  double rx, ry, rz;
		  mousePoleGo(r1x,r1y,r1z, r2x,r2y,r2z, 
					  static_cast<double>(theCubes[selectedCube].x), 
					  static_cast<double>(theCubes[selectedCube].y),
					  static_cast<double>(theCubes[selectedCube].z),
					  rx, ry, rz,
					  (Fl::event_state() & FL_CTRL) != 0);
		  theCubes[selectedCube].x = (float) rx;
		  theCubes[selectedCube].y = (float) ry;
		  theCubes[selectedCube].z = (float) rz;
		  damage(1);
		};
		break;
	  }
	  return 1;
  
  case FL_KEYBOARD:
	  return 0;

  default:
	return 0;
  }
}

//**************************************************************************
//
// this has to be a method since we need access to make_current
//==========================================================================
void MyGlWindow::doPick()
//==========================================================================
{
	int mx = Fl::event_x(); // where is the mouse?
	int my = Fl::event_y();
	make_current();		// so we can do drawing outside of
						// draw method
	// we need the size of the OpenGL window, in its own terms
	int viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	// set up the pick matrix on the stack - remember, FlTk is
	// upside down!
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity ();
	gluPickMatrix((double)mx, (double)(viewport[3]-my), 5, 5, viewport);
	// now, setup the drawing transforms
	setProjection(0);
	// now draw the objects - but really only see what we hit
	GLuint buf[100];
	glSelectBuffer(100,buf);
	glRenderMode(GL_SELECT);
	glInitNames();
	glPushName(0);
	drawCubes(0,1);
	// go back to drawing mode, and see how picking did
	int hits = glRenderMode(GL_RENDER);
	if (hits) {
		// warning; this just grabs the first object hit - if there
		// are multiple objects, you really want to pick the closest
		// one - see the OpenGL manual 
		// remember: we load names that are one more than the index
		selectedCube = buf[3]-1;
	} else // nothing hit, nothing selected
		selectedCube = -1;
}

//**************************************************************************
//
// get the mouse in NDC
//==========================================================================
void MyGlWindow::getMouseNDC(float& x, float& y)
//==========================================================================
{
	// notice, we put everything into doubles so we can do the math
	  float mx = (float) Fl::event_x();	// remeber where the mouse went down
	  float my = (float) Fl::event_y();

	  // we will assume that the viewport is the same as the window size
	  float wd = (float) w();
	  float hd = (float) h();

	  // remember that FlTk has Y going the wrong way!
	  my = hd-my;
	
	  x = (mx / wd) * 2.0f - 1.f;
	  y = (my / hd) * 2.0f - 1.f;
}

// the main routine makes the window, and then runs an even loop
// until the window is closed
int main(int, char**)
{
	Fl_Double_Window* wind = new Fl_Double_Window(100,100,600,400,"GL 3D Sample");

	wind->begin();		// put widgets inside of the window
		MyGlWindow* gl = new MyGlWindow(10,10,380,380);
		Fl_Multiline_Output* out = new Fl_Multiline_Output(420,140,580-420,380-140);
		out->value("CubeIt 2! - 3D UI\n\nLeftMouse: select&drag\nRightMouse: ArcBall\nALT+Right Mouse: pan camera\n\nDrag parallel to floor\nHold CTRL for vertical");
	wind->end();


	wind->show();	// this actually opens the window

	// make some cubes to show
	theCubes.push_back(Cube(-3,3, 3));
	theCubes.push_back(Cube( 3,3, 3));
	theCubes.push_back(Cube(-3,3,-3));
	theCubes.push_back(Cube( 3,3,-3));
	// put some cubes over space so we can test the shadows only cast
	// onto the floor (not onto space)
	theCubes.push_back(Cube(7,3,0,1));
	theCubes.push_back(Cube(-7,3,0,1));
	theCubes.push_back(Cube(0,3,7,1));
	theCubes.push_back(Cube(0,3,-7,1));
	// put a cube over another cube so we can see that the shadows
	// don't stack
	theCubes.push_back(Cube( 0,4,-4,2));
	theCubes.push_back(Cube( 0,2,-3,1));

	// become interactive
	Fl::run();
	delete wind;

	return 1;
}

